import { Pool, PoolClient, Query, QueryResult, QueryResultRow } from 'pg';
import { loadConfig } from './configLoader';
import { unifyArrays, randomArray } from './helperFunctions'
import fs from 'fs';
const config = loadConfig();

class SqlSystem {

  private static pool: Pool;


  public static async InitDB() {
    this.pool = new Pool({
      database: 'postgres',
      host: process.env.DB_HOST,
      user: process.env.DB_USER,
      password: process.env.DB_PASSWORD,
      max: config.data.worker_count ?? 10
    });
    let conn: PoolClient | undefined;
    try {
      conn = await this.pool.connect();
      try {
        await conn.query('BEGIN');
        let qry: string = fs.readFileSync('/app/sql/init.sql', 'utf8');
        await conn.query(qry);
        await conn.query('COMMIT');
      } catch (error) {
        console.error("InitDB Error: ", error);
        if (conn) await conn.query('ROLLBACK');
        if (conn) conn.release();
        this.pool.end();
        throw error;
      }

      //await this.TimeoutThread(10000)

      await conn.query('BEGIN');

      console.log("Inserting random numbers into Pool SETUP");
      let dummy_auctionid: string[] = randomArray(10, 8), dummy_lbin: number[] = randomArray(10, 5);
      await this.Upsert(dummy_auctionid, dummy_lbin);

      await conn.query('COMMIT');

      console.log("Requesting All from table auctions PRE REMOVE");
      console.log(await this.Query("SELECT * FROM auctions", false));

      await conn.query('BEGIN');

      await this.RemoveAll()

      await conn.query('COMMIT')

      console.log("Requesting All from table auctions POST REMOVE");
      console.log(await this.Query("SELECT * FROM auctions", false));
    }
    catch (error) {
      console.error("InitDB Error: ", error);
      if (conn) await conn.query('ROLLBACK');
      throw error;
    }
    finally {
      if (conn) conn.release();
    }
  }
  public static async Upsert(auctionid: string | string[], lbin: number | number[]) {
    let conn: PoolClient | undefined;
    try {
      conn = await this.pool.connect();
      await conn.query('BEGIN');
      if (Array.isArray(auctionid) && Array.isArray(lbin))
        if (auctionid.length === lbin.length) {
          await conn.query(`
            INSERT INTO auctions (id, auctionid, lbin)
              SELECT uuid_generate_v4()::UUID, unnest($1::text[]), unnest($2::decimal[])
              ON CONFLICT (auctionid)
                DO UPDATE SET lbin = EXCLUDED.lbin;
          `, [auctionid, lbin]);
        }
        else if (Array.isArray(auctionid) || Array.isArray(lbin))
          throw Error(`Upsert SQL Function error - cannot unify collection ${Array.isArray(auctionid) ? auctionid : lbin} to singlet ${!Array.isArray(auctionid) ? auctionid : lbin}`);
        else {
          await conn.query(`
          INSERT INTO auctions (id, auctionid, lbin) 
          VALUES uuid_generate_v4()::UUID, $1::text, $2::decimal
            ON CONFLICT (auctionid)
              DO UPDATE SET lbin = EXCLUDED.lbin;
        `, [auctionid, lbin]);
        }

      await conn.query('COMMIT');
    }
    catch (error) {
      console.error("message_before_error: ", error);
      if (conn) await conn.query('ROLLBACK');
      throw error;
    }
    finally {
      if (conn) conn.release();
    }
  }
  //REMOVE ALL ELEMENTS IN AUCTIONS TABLE 
  public static async RemoveAll() {
    let conn: PoolClient | undefined;
    try {
      conn = await this.pool.connect();
      await conn.query('BEGIN');
      await conn.query('DELETE FROM auctions')
      await conn.query('COMMIT');
    }
    catch (error) {
      console.error("InitDB Error: ", error);
      if (conn) await conn.query('ROLLBACK');
      throw error;
    }
    finally {
      if (conn) conn.release();
    }
  }
  //REMOVE ELEMENT IN AUCTIONS TABLE BY AUCTION-ID
  public static async RemoveByAuctionId(auctionid: string | string[]) {
    let conn: PoolClient | undefined;
    try {
      conn = await this.pool.connect();
      await conn.query('BEGIN');
      if (Array.isArray(auctionid)) {
        await conn.query('DELETE FROM auctions WHERE auctionid IN ($1);', [auctionid]);
      }
      else {
        await conn.query('DELETE FROM auctions WHERE auctionid = $1;', [auctionid]);
      }
      await conn.query('COMMIT');
    }
    catch (error) {
      console.error("InitDB Error: ", error);
      if (conn) await conn.query('ROLLBACK');
      throw error;
    }
    finally {
      if (conn) conn.release();
    }
  }

  public static async Query(query: string, IsWriting: boolean = false) {
    let conn: PoolClient | undefined;
    let result: any;
    try {
      conn = await this.pool.connect()
      if (IsWriting) {
        await conn.query('BEGIN');
        conn.query(query);
        await conn.query('COMMIT');
      }
      else {
        console.log('TRYING TO QUERY');
        result = (await conn.query(query)).rows;
      }
    }
    catch (error) {
      console.error("InitDB Error: ", error);
      if (IsWriting && conn) await conn.query('ROLLBACK');
      throw error;
    }
    finally {
      if (conn) conn.release();
      if (!IsWriting) return result;
    }
  }
  ////////////////////////////////////////////////////////////////////////////////////////////////////// NOT FIXED / UNTESTED FUNCTIONS
  //MATCH PROVIDED ELEMENTS IN AUCTIONS TABLE - returns true/false
  public static async Match(auctionid: string | string[], lbin: number | number[]): Promise<boolean | any> {
    let conn: PoolClient | undefined;
    let result;
    try {
      conn = await this.pool.connect();
      if (Array.isArray(auctionid) && Array.isArray(lbin))
        if (auctionid.length === lbin.length) {
          result = await conn.query(`
            CREATE TEMPORARY TABLE IF NOT EXISTS TEMP (
            auctionID VARCHAR(255) PRIMARY KEY, 
            lbin DECIMAL(65,5)
            );
            INSERT INTO TEMP (auctionID, lbin) VALUES ?
            SELECT tmp.*
              FROM TEMP AS tmp
              WHERE NOT EXISTS (
              SELECT 1 
                FROM auctions AS auc
                WHERE auc.auctionID = tmp.auctionID
                AND auc.lbin = tmp.lbin);
            `, [await unifyArrays(auctionid, lbin)]);
          return result.rows;
        }
        else if (Array.isArray(auctionid) || Array.isArray(lbin))
          throw Error(`Match SQL Function error - cannot unify collection ${Array.isArray(auctionid) ? auctionid : lbin} to singlet ${!Array.isArray(auctionid) ? auctionid : lbin}`);
        else {
          result = await conn.query(`
          SELECT CASE 
            WHEN COUNT(*) > 0 
              THEN true 
              ELSE false 
            END AS user_exists 
          FROM auctions WHERE auctionID = ?;
        `, [auctionid]);
          return Boolean(result.rows.length);
        }
    }
    catch (error) {
      console.error("message_before_error: ", error);
      throw error;
    }
    finally {
      if (conn) conn.release();
    }
  }
  //INSERT ELEMENT IN AUCTIONS TABLE OR UPDATE IF IT ALREADY EXISTS
  public static async deprecated_Upsert(auctionid: string | string[], lbin: number | number[]) {
    let conn: PoolClient | undefined;
    try {
      conn = await this.pool.connect();
      await conn.query('BEGIN');
      if (Array.isArray(auctionid) && Array.isArray(lbin))
        if (auctionid.length === lbin.length) {
          await conn.query(`
            INSERT INTO auctions (auctionid, LBin)
            VALUES ($1,$2)
            ON CONFLICT (auctionid)
            DO UPDATE SET LBin = $2;
          `, [auctionid, lbin]);
        }
        else if (Array.isArray(auctionid) || Array.isArray(lbin))
          throw Error(`Upsert SQL Function error - cannot unify collection ${Array.isArray(auctionid) ? auctionid : lbin} to singlet ${!Array.isArray(auctionid) ? auctionid : lbin}`);
        else {
          await conn.query(`
          INSERT INTO auctions (auctionid, LBin)
          VALUES ($1,$2)
          ON CONFLICT (auctionid)
          DO UPDATE SET LBin = $2;
        `, [auctionid, lbin]);
        }
      await conn.query('COMMIT');
    }
    catch (error) {
      console.error("message_before_error: ", error);
      if (conn) await conn.query('ROLLBACK');
      throw error;
    }
    finally {
      if (conn) conn.release();
    }
  }
}
export {
  SqlSystem
}
